/*  GnuPOSkeyboard.c                      */
/*  This contains the keyboard interface  */
/*  routines for GnuPOS, including the    */
/*  reading and writing of the xml state  */
/*  definitions file.                     */
/*                                        */
/*  (c) 2001 Conrad Canterford.           */

#include <gnome.h>
#include <gnome-xml/xmlmemory.h>
#include <gnome-xml/parser.h>

#include "GnuPOS.h"
#include "GnuPOS-P.h"
#include "GnuPOS-keyboard-P.h"
#include "GnuPOS-keymap-defines.h"

state_master_struct all_states;

/* Table holding symbol-value pairs for */
/*  keyboard codes                      */
symval_struct       *keysym_table;
guint                keysym_table_count;

/* Table holding symbol-value pairs for */
/* functions                            */
functions_struct    *function_table;
guint                function_table_count;

/* Table holding symbol-value pairs for */
/* parameters (registers)               */
registers_struct    *register_table;
guint                register_table_count;

extern gnp_global_data_struct gnp_global_data;



/*                                         */
/*       KEYMAP ROUTINES                   */
/*                                         */
/* Group together the routines for reading */
/* the keymap and for processing keystrokes*/

void gnp_initialise_tables()
{
  all_states.current_state=0;
  gnp_read_keysyms();
  gnp_read_keymaps();
}

void gnp_initialise_functions()

/* Loads the inbuilt functions into the */
/* function names table.                */

{
  functions_struct    init[]={
                           {"reset",    GNPF_RESET,0,NULL,gnp_reset_func,GNPS_UNDEF},
                           {"clear",    GNPF_CLEAR,1,NULL,gnp_clear_func,GNPS_UNDEF},
                           {"add",      GNPF_ADD,3,NULL,NULL,GNPS_UNDEF},
                           {"subtract", GNPF_SUBTRACT,3,NULL,NULL,GNPS_UNDEF},
                           {"multiply", GNPF_MULTIPLY,3,NULL,NULL,GNPS_UNDEF},
                           {"remainder",GNPF_REMAINDER,3,NULL,NULL,GNPS_UNDEF},
                           {"move",     GNPF_MOVE,2,NULL,NULL,GNPS_UNDEF},
                           {"display",  GNPF_DISPLAY,2,NULL,NULL,GNPS_UNDEF},
                           {"output",   GNPF_OUTPUT,2,NULL,NULL,GNPS_UNDEF},
                           {"quit",     GNPF_QUIT,0,NULL,NULL,GNPS_UNDEF},
                           {"divide",   GNPF_DIVIDE,3,NULL,NULL,GNPS_UNDEF}
                          };

  if(gnp_global_data.verbose_status)
    printf("  Commencing copy of inbuilt functions table\n");
  memcpy(function_table,init,sizeof(functions_struct)*GNP_INTERNAL_FUNCTION_COUNT);  
  if(gnp_global_data.verbose_status)
    printf("  Completed copy of inbuilt functions table\n");
}

void gnp_initialise_registers()

/* Loads the predefined registers into the */
/* register table.                         */

{
  registers_struct  init[]={
                         {"key",       GNPR_KEY,       GNP_REGISTER_SPECIAL,GNP_VALUE_UNDEF,0},
                         {"return",    GNPR_RETURN,    GNP_REGISTER_SPECIAL,GNP_VALUE_UNDEF,0},
                         {"statename", GNPR_STATENAME, GNP_REGISTER_SPECIAL,GNP_VALUE_STRING,NULL},
                         {"",          GNPR_STRING,    GNP_REGISTER_PRIVATE,GNP_VALUE_STRING,NULL},
                         {"",          GNPR_NUMBER,    GNP_REGISTER_PRIVATE,GNP_VALUE_INTEGER,0},
                         {"mode",      GNPR_MODETEXT,  GNP_REGISTER_DISPLAY,GNP_VALUE_STRING,NULL},
                         {"display",   GNPR_DISPLAY,   GNP_REGISTER_DISPLAY,GNP_VALUE_STRING,NULL},
                         {"helptext",  GNPR_HELPTEXT,  GNP_REGISTER_DISPLAY,GNP_VALUE_STRING,NULL},
                         {"logtext",   GNPR_LOGTEXT,   GNP_REGISTER_DISPLAY,GNP_VALUE_STRING,NULL},
                         {"statustext",GNPR_STATUSTEXT,GNP_REGISTER_DISPLAY,GNP_VALUE_STRING,NULL}
                        };

  if(gnp_global_data.verbose_status)
    printf("  Commencing copy if internal registers table\n");
  memcpy(register_table,init,sizeof(registers_struct)*GNP_INTERNAL_REGISTER_COUNT);
  if(gnp_global_data.verbose_status)
    printf("  Completed copy of internal registers table\n");
}

void gnp_read_keysyms()

/* Reads the keyboard value <--> symbol    */
/* lookup table from the keysym xml file.  */

{
  xmlDocPtr  document;
  xmlNodePtr node,node_store,np;
  xmlNsPtr   namespace;
  guint      countmaps=0;
  guint      i,j;
  gboolean   has_sym,has_code;
  gchar     *thisname;

  if(gnp_global_data.verbose_status)
    printf("Reading keysyms file\n");

  document=xmlParseFile("/home/conrad/Projects/gnuPOS/keysyms.xml");
  if(document==NULL)
  {
    printf("ERROR: Cannot read keysyms file. ABORTING.\n");
    exit(-1);
  }

  node=xmlDocGetRootElement(document);
  if(node==NULL)
  {
    printf("ERROR: keysyms file is empty. ABORTING.\n");
    exit(-1);
  }

  namespace=xmlSearchNsByHref(document,node,(const xmlChar *) "http://www.gnucash.org/keysyms/1.0");
  if(namespace==NULL)
  {
    printf("ERROR: keysyms file of wrong type. ABORTING.\n");
    exit(-1);
  }
  if (xmlStrcmp(node->name, (const xmlChar *) "keysyms"))
  {
    printf("ERROR: GnuPOS data not found in file. ABORTING.\n");
    exit(-1);
  }

  node=node->xmlChildrenNode;
  node_store=node;

  while(node!=NULL)
  {
    /* count how many, so we know how big to make the keysyms table */
    if((!xmlStrcmp(node->name,(const xmlChar *)"map"))&&
      (node->ns==namespace))
    {
      countmaps++;
    }
    node=node->next;
  }

  if(gnp_global_data.verbose_status)
  {
    printf("Read %d Keyboard maps\n",countmaps);
    printf("Will add %d special maps as well.\n",GNP_INTERNAL_KEY_COUNT);
  }

  countmaps+=GNP_INTERNAL_KEY_COUNT;

  keysym_table=(symval_struct *)g_malloc(sizeof(symval_struct)*countmaps);
  if(keysym_table==NULL)
  {
    printf("ERROR: Unable to get memory for keysyms table. ABORTING\n");
    exit(-1);
  }

  keysym_table_count=countmaps;

  countmaps=0;

  keysym_table[countmaps].sym="%numeric";
  keysym_table[countmaps].code=GNPK_NUMERIC;
  countmaps++;
  keysym_table[countmaps].sym="%lowercase";
  keysym_table[countmaps].code=GNPK_LOWERCASE;
  countmaps++;
  keysym_table[countmaps].sym="%uppercase";
  keysym_table[countmaps].code=GNPK_UPPERCASE;
  countmaps++;
  keysym_table[countmaps].sym="%alpha";
  keysym_table[countmaps].code=GNPK_ALPHA;
  countmaps++;
  keysym_table[countmaps].sym="%alphnum";
  keysym_table[countmaps].code=GNPK_ALPHNUM;
  countmaps++;
  keysym_table[countmaps].sym="%punct";
  keysym_table[countmaps].code=GNPK_PUNCT;
  countmaps++;
  keysym_table[countmaps].sym="%spaces";
  keysym_table[countmaps].code=GNPK_SPACES;
  countmaps++;
  keysym_table[countmaps].sym="%printable";
  keysym_table[countmaps].code=GNPK_PRINTING;
  countmaps++;
  keysym_table[countmaps].sym="%special";
  keysym_table[countmaps].code=GNPK_CONTROL;
  countmaps++;
  keysym_table[countmaps].sym="%default";
  keysym_table[countmaps].code=GNPK_DEFAULT;
  countmaps++;

  node=node_store;
  while(node!=NULL)
  {
    /* read each map definition and put it into the keysyms table */
    if((!xmlStrcmp(node->name,(const xmlChar *)"map"))&&
      (node->ns==namespace))
    {
      np=node->xmlChildrenNode;

      /* There are, and must always be, precisely 2 */
      /* children for each node. If not, xml file   */
      /* is malformed and should be exterminated.   */
      has_sym=FALSE;
      has_code=FALSE;
      for(i=0;i<2;i++)
      {
        if(np==NULL)
        {
          printf("ERROR: <map> entry %d does not have both a <sym> and <code> entry. ABORTING\n",countmaps);
          exit(-1);
        }
        if((!xmlStrcmp(np->name,(const xmlChar *)"sym")) &&
             (np->ns == namespace))
        {
          /* Check this is not a duplicate  */
          thisname=xmlNodeListGetString(document,np->xmlChildrenNode, 1);
          j=0;
          while(j<countmaps)
          {
            if(strcmp(thisname,keysym_table[j].sym)==0)
            {
              printf("ERROR: Duplicate key symbol entry in <map> entry %d. ABORTING\n",countmaps+1);
              exit(-1);
            }
            j++;
          }
          keysym_table[countmaps].sym=thisname;           
          has_sym=TRUE;
        }
        else
        {
          if((!xmlStrcmp(np->name,(const xmlChar *)"code")) &&
               (np->ns == namespace))
          {
            keysym_table[countmaps].code=
               strtol(xmlNodeListGetString(document,np->xmlChildrenNode, 1),NULL,0);
            has_code=TRUE;
          }
        }
        np=np->next;
      }
      if(!has_sym||!has_code)
      {
        printf("ERROR: <map> entry %d does not have both a <sym> and <code> entry. ABORTING\n",countmaps);
        exit(-1);
      }
      if(keysym_table[countmaps].code==0xffffff)
      {
        printf("ERROR: Reserved code 0xffffff used in <map> entry %d. ABORTING\n",countmaps);
        exit(-1);
      }
      countmaps++;
    }
    node=node->next;
  }
  if(gnp_global_data.verbose_status)
  {
    int i;

    printf("Keysym table output:\n");
    for(i=0;i<countmaps;i++)
      printf("%d: %s -> %x\n",i,keysym_table[i].sym,keysym_table[i].code);
  }
}


void gnp_read_keymaps()

/* Read the keymaps XML file as pointed to */
/* by our config, and create the keymaps   */
/* memory structures used to guide input.  */

{
  xmlDocPtr  document;
  xmlNodePtr node,rootnode,innerstates;
  xmlNodePtr initnode=NULL;
  xmlNsPtr   namespace;
  guint      countstates=0,i=0;
  guint      countfuncs=0,countregs=0;
  guint      initialstate;
  gchar     *thisname;

  if(gnp_global_data.verbose_status)
    printf("Reading keymap file\n");

  document=xmlParseFile("/home/conrad/Projects/gnuPOS/samplekeymap.xml");
  if(document==NULL)
  {
    printf("ERROR: Cannot read keymap file. ABORTING.\n");
    exit(-1);
  }

  node=xmlDocGetRootElement(document);
  if(node==NULL)
  {
    printf("ERROR: keymap file is empty. ABORTING.\n");
    exit(-1);
  }

  namespace=xmlSearchNsByHref(document,node,(const xmlChar *) "http://www.gnucash.org/gnuPOS/1.0");
  if(namespace==NULL)
  {
    printf("ERROR: keymap file of wrong type. ABORTING.\n");
    exit(-1);
  }
  if (xmlStrcmp(node->name, (const xmlChar *) "GnuPOS"))
  {
    printf("ERROR: GnuPOS data not found in file. ABORTING.\n");
    exit(-1);
  }

  rootnode=node;
  node=node->xmlChildrenNode;

  /* The tree is parsed in three passes. In the first    */
  /* pass we parse all the definitions, but only count   */
  /* the number of states. After that pass is completed, */
  /* we can grab enough memory for the state array, and  */
  /* we can parse the states confident that all the      */
  /* definitions have been processed. The second pass    */
  /* inserts all the state names into names table, so    */
  /* that, on the third pass, we can correctly process   */
  /* all input definitions.                              */

  /* Pass 1: Count all bits  */
  if(gnp_global_data.verbose_status)
    printf("  Commencing Pass 1\n");
  while(node!=NULL)
  {
    if(node&&xmlIsBlankNode(node))
    {
      /* printf("Blank node - name: %s, namespace %s\n",node->name, node->ns); */
      node=node->next;
    }
    else
      if((!xmlStrcmp(node->name,(const xmlChar *)"definitions"))&&(node->ns== namespace))
      {
        /* at this point, we only count instances */
        innerstates=node->xmlChildrenNode;
        while(innerstates!=NULL)
        {
          if((!xmlStrcmp(innerstates->name,(const xmlChar *)"function"))&&
             (innerstates->ns==namespace))
            countfuncs++;
          if((!xmlStrcmp(innerstates->name,(const xmlChar *)"register"))&&
             (innerstates->ns==namespace))
            countregs++;
          innerstates=innerstates->next;
        }
        node=node->next;
      }
      else
        if((!xmlStrcmp(node->name,(const xmlChar *)"states")) &&
            (node->ns == namespace))
        {
          /* Don't parse state tree here, just count them */
          innerstates=node->xmlChildrenNode;
          while(innerstates!=NULL)
          {
            if((!xmlStrcmp(innerstates->name,(const xmlChar *)"setstate"))&&
              (innerstates->ns==namespace))
            {
              countstates++;
              innerstates=innerstates->next;
            }
          }
          node=node->next;
        }
        else
          if((!xmlStrcmp(node->name,(const xmlChar *)"initialisation"))&&(node->ns== namespace))
          {
            /*  we only store the node pointer here  */
            /* we parse this later when the states   */
            /* tree has been built, so we can set the*/
            /* initial state properly.               */

            if(initnode==NULL)
              initnode=node;
            else
            {
              printf("ERROR: Mulitple initialisation nodes! ABORTING\n");
              exit(-1);
            }
            node=node->next;
          }
          else
          {
             printf("ERROR: Unknown node name %s\n",node->name);
            printf("   %s ignored.\n",node->name);
            node=node->next;
          }
  }

  /*  grab memory for states array and names table */
  if(gnp_global_data.verbose_status)
    printf("    States read: %d, Functions read: %d, Registers read: %d\n",countstates,countfuncs,countregs);
  
  all_states.states=g_malloc(sizeof(state_instance_struct)*countstates);
  if(all_states.states==NULL)
  {
    printf("ERROR: Could not allocate memory for state array.\n");
    exit(-1);
  }
  all_states.nametable=g_malloc(sizeof(gchar *)*countstates);
  if(all_states.nametable==NULL)
  {
    printf("ERROR: Could not allocate memory for nametable array.\n");
    exit(-1);
  }
  all_states.max_states=countstates;


  /* Now modify the function count to include the   */
  /* built-in functions, as these are stored in the */
  /* functions name lookup table.                   */
  countfuncs+=GNP_INTERNAL_FUNCTION_COUNT;
  function_table=(functions_struct *)g_malloc(sizeof(functions_struct)*countfuncs);
  if(function_table==NULL)
  {
    printf("ERROR: Unable to get memory for function table. ABORTING\n");
    exit(-1);
  }
  function_table_count=countfuncs;

  countregs+=GNP_INTERNAL_REGISTER_COUNT;
  register_table=(registers_struct *)g_malloc(sizeof(registers_struct)*countregs);
  if(register_table==NULL)
  {
    printf("ERROR: Unable to get memory for register table. ABORTING\n");
    exit(-1);
  }
  register_table_count=countregs;

  /* This must be done here rather than earlier, so that */
  /* the tables have been created with the right size.   */
  gnp_initialise_functions();
  gnp_initialise_registers();
  
  if(gnp_global_data.verbose_status)
    printf("  Commencing Pass 2\n");

  /* Pass 2: Read state, function and register names */

  /* Do NOT, I repeat DO NOT free the memory allocated   */
  /* to "thisname" during this pass! This memory         */
  /* contains the actual strings which are pointed to    */
  /* by the tables.                                      */

  countstates=0;
  countfuncs=GNP_INTERNAL_FUNCTION_COUNT;
  countregs=GNP_INTERNAL_REGISTER_COUNT;

  node=rootnode->xmlChildrenNode;
  while(node!=NULL)
  {
    if((!xmlStrcmp(node->name,(const xmlChar *)"states")) &&
        (node->ns == namespace))
    {
      if(gnp_global_data.verbose_status)
        printf("  Parsing <state> (pass 2)\n");


      /* This time, we parse the state tree and extract names only */
      innerstates=node->xmlChildrenNode;
      while(innerstates!=NULL)
      {
        if((!xmlStrcmp(innerstates->name,(const xmlChar *)"setstate"))&&
          (innerstates->ns==namespace))
        {
          /*  read name of state (its a property of node) */
          /*  and insert that into the state name lookup  */
          /*  table unless it is there already. If it is  */
          /*  a duplicate, print an error and kill it.    */

          thisname=(gchar *)g_malloc(strlen(xmlGetProp(innerstates,(const xmlChar *)"name")));
          strcpy(thisname,xmlGetProp(innerstates,(const xmlChar *)"name"));
          if(countstates>0)
          {
            for(i=0;i<countstates;i++)
            {
              if(strcmp(thisname,(char *)all_states.nametable[i])==0)
              {
                printf("ERROR: Duplicate state name %s\n",thisname);
                exit(-1);
              }
            }
          }
          all_states.nametable[countstates]=thisname;
          countstates++;
        }
        innerstates=innerstates->next;
      }
    }
    if((!xmlStrcmp(node->name,(const xmlChar *)"definitions")) &&
        (node->ns == namespace))
    {
      /* Parse the definitions tree and extract names only */
      innerstates=node->xmlChildrenNode;
      while(innerstates!=NULL)
      {
        if((!xmlStrcmp(innerstates->name,(const xmlChar *)"function"))&&
          (innerstates->ns==namespace))
        {
          /*  read the name of the function (its a        */
          /*  property of the node) and insert that name  */
          /*  into the function name lookup table (unless */
          /*  it is there already). If it is a duplicate, */
          /*  print an error then kill it.                */

          thisname=(gchar *)g_malloc(strlen(xmlGetProp(innerstates,(const xmlChar *)"name")));
          strcpy(thisname,xmlGetProp(innerstates,(const xmlChar *)"name"));

          if(gnp_global_data.verbose_status)
            printf("    Processing function %s.\n",thisname);

          for(i=0;i<countfuncs;i++)
          {
            if(strcmp(thisname,(char *)function_table[i].sym)==0)
            {
              printf("ERROR: Duplicate function name %s\n",thisname);
              exit(-1);
            }
          }
          function_table[countfuncs].sym=thisname;
          function_table[countfuncs].code=countfuncs-GNP_INTERNAL_FUNCTION_COUNT;
          countfuncs++;
        }
        if((!xmlStrcmp(innerstates->name,(const xmlChar *)"register"))&&
          (innerstates->ns==namespace))
        {
          /*  read name of the register (its a property   */
          /*  of the node) and insert that into the       */
          /*  register name lookup table (unless it is    */
          /*  there already. If it is a duplicate, print  */
          /*  an error and kill it.                       */
          thisname=(gchar *)g_malloc(strlen(xmlGetProp(innerstates,(const xmlChar *)"name")));
          strcpy(thisname,xmlGetProp(innerstates,(const xmlChar *)"name"));

          if(gnp_global_data.verbose_status)
            printf("    Processing register %s.\n",thisname);

          if(countregs>0)
          {
            for(i=0;i<countregs;i++)
            {
              if(strcmp(thisname,(char *)register_table[i].sym)==0)
              {
                printf("ERROR: Duplicate register name %s\n",thisname);
                exit(-1);
              }
            }
          }
          register_table[countregs].sym=thisname;
          register_table[countregs].code=countregs-GNP_INTERNAL_REGISTER_COUNT;
          /*  TODO: read this stuff from the file!  */
          register_table[countregs].type=GNP_REGISTER_NORMAL;
          register_table[countregs].value_type=GNP_VALUE_UNDEF;
          register_table[countregs].value=0;
          countregs++;
        }
        innerstates=innerstates->next;
      }
    }
    node=node->next;
  }

  if(gnp_global_data.verbose_status)
    printf("  Commencing Pass 3\n");

  /* Pass 3: Parse everything properly  */
  /* Because all the names have been parsed */
  /* already, it now doesn't matter that    */
  /* some of these will refer to other stuff*/
  /* that hasn't been parsed yet. The       */
  /* codes will be substituted, and found at*/
  /* run-time.                              */

  countstates=0;
  countfuncs=GNP_INTERNAL_FUNCTION_COUNT;
  node=rootnode->xmlChildrenNode;
  while(node!=NULL)
  {
    if((!xmlStrcmp(node->name,(const xmlChar *)"states")) &&
        (node->ns == namespace))
    {
      /* This time, we parse the state tree completely   */
      innerstates=node->xmlChildrenNode;
      while(innerstates!=NULL)
      {
        if((!xmlStrcmp(innerstates->name,(const xmlChar *)"setstate"))&&
          (innerstates->ns==namespace))
        {
          if(gnp_parse_state(document,innerstates,countstates))
            countstates++;
        }
        innerstates=innerstates->next;
      }
    }
    else
      if((!xmlStrcmp(node->name,(const xmlChar *)"definitions")) &&
          (node->ns == namespace))
      {
        /* This time, we parse the state tree completely   */
        innerstates=node->xmlChildrenNode;
        while(innerstates!=NULL)
        {
          if((!xmlStrcmp(innerstates->name,(const xmlChar *)"function"))&&
            (innerstates->ns==namespace))
          {
            gnp_parse_function(document,innerstates,countfuncs); 
            countfuncs++;
          }
          innerstates=innerstates->next;
        }
      }
    node=node->next;
  }

  /*  parse initialisation node  */
  /* TODO: DO THIS!! Currently defaults to 0!  */
  initialstate=0;

  gnp_change_state(initialstate);

  if(gnp_global_data.verbose_status)
  {
    int j;
    commands_struct *ptr;

    printf("\nRegister table output:\n");

    for(i=0;i<register_table_count;i++)
    {
      printf("-------------------------------------------------------\n");
      printf("%d: %s -> %x\n",i,register_table[i].sym,register_table[i].code);
    }

    printf("\nFunction table output:\n");

    for(i=0;i<function_table_count;i++)
    {
      printf("-------------------------------------------------------\n");
      printf("%d: %s -> %x\n",i,function_table[i].sym,function_table[i].code);
      ptr=function_table[i].command;
      while(ptr!=NULL)
      {
        printf("         func: %x (%s), par1: %x (%s), par2: %x (%s), par3: %x (%s)\n",
           ptr->function,gnp_lookup_function_name(ptr->function),ptr->param1,gnp_lookup_parameter_name(ptr->param1),
           ptr->param2,gnp_lookup_parameter_name(ptr->param2),ptr->param3,gnp_lookup_parameter_name(ptr->param3));
        ptr=ptr->next;
      }
    }
    printf("\nState table output:\n");

    for(i=0;i<all_states.max_states;i++)
    {
      printf("-------------------------------------------------------\n");
      printf("%d: %s",i,all_states.nametable[i]);
      printf(" - Help-text: %s\n",all_states.states[i].helptext);
      printf("   Inputs defined: %d -\n",all_states.states[i].numinputs);
      for(j=0;j<all_states.states[i].numinputs;j++)
      {
        printf("      %d: key: %x (%s) endstate: %d (%s)\n",
           j,all_states.states[i].inputs[j].inpchar,gnp_lookup_key_name(all_states.states[i].inputs[j].inpchar),
           all_states.states[i].inputs[j].endstate,gnp_lookup_state_name(all_states.states[i].inputs[j].endstate));
        /* print of commands to go here */
        ptr=all_states.states[i].inputs[j].command;
        while(ptr!=NULL)
        {
          printf("         func: %x (%s), par1: %x (%s), par2: %x (%s), par3: %x (%s)\n",
             ptr->function,gnp_lookup_function_name(ptr->function),ptr->param1,gnp_lookup_parameter_name(ptr->param1),
             ptr->param2,gnp_lookup_parameter_name(ptr->param2),ptr->param3,gnp_lookup_parameter_name(ptr->param3));
          ptr=ptr->next;
        }
      }
    }
  }

  printf("Initial state: %d (%s)\n",initialstate,gnp_lookup_state_name(initialstate));
}

gboolean gnp_parse_state(xmlDocPtr doc, xmlNodePtr node, guint state)

/*  Do all the muck associated with building */
/* a complete state array entry for a state. */
/* node is the parent node of this state.    */
/* state is the array entry counter for the  */
/*     state array.                          */

{
  xmlNodePtr  np1,np2;
  guint       i,lookup;
  gchar       *thisname;
  gchar       *performstring;
  guint       countinputs=0;

  if(gnp_global_data.verbose_status)
    printf("Processing state %s\n",all_states.nametable[state]);
  
  np1=node->xmlChildrenNode;

  /* Initialise this to NULL here, in case no helptext */
  /* is found in the xml file.                         */
  all_states.states[state].helptext=NULL;

  while(np1!=NULL)
  {
    if(!xmlStrcmp(np1->name,(const xmlChar *)"input"))
    {
      /* Once again, only count them at this stage */
      countinputs++;
    }
    else
      if(!xmlStrcmp(np1->name,(const xmlChar *)"helptext"))
      {
        /* process helptext definition */
        all_states.states[state].helptext=xmlNodeListGetString(doc,np1->xmlChildrenNode, 1);
      }
    np1=np1->next;
  }
  /* allocate memory, then rescan all the inputs */
  all_states.states[state].inputs=(state_instance_struct *)g_malloc(sizeof(state_instance_struct)*countinputs);
  all_states.states[state].numinputs=countinputs;

  if(gnp_global_data.verbose_status)
    printf("  Read %d inputs\n",countinputs);

  countinputs=0;

  /*  Now, we rescan the node, processing  */
  /* input nodes this time.                */

  np1=node->xmlChildrenNode;

  while(np1!=NULL)
  {
    if(!xmlStrcmp(np1->name,(const xmlChar *)"input"))
    {
      /* process input definition */
      /* Start by getting the key this applies */
      /* to - its a property of the input node.*/
      thisname=(gchar *)g_malloc(strlen(xmlGetProp(np1,(const xmlChar *)"key")));
      strcpy(thisname,xmlGetProp(np1,(const xmlChar *)"key"));

      if(gnp_global_data.verbose_status)
        printf("  Processing input %s\n",thisname);

      lookup=gnp_lookup_key(thisname);
      if(lookup==0xffffff)
        printf("WARNING: Undefined key \"%s\" in state \"%s\". Ignored.\n",thisname,all_states.nametable[state]);
      all_states.states[state].inputs[countinputs].inpchar=lookup;

      if(gnp_global_data.verbose_status)
        printf("    Input key translated to %x\n",lookup);

      /* Now, read the child nodes to retreive */
      /* perform code and endstate             */

      np2=np1->xmlChildrenNode;

      /* Flag the endstate with the special state */
      /* in case an endstate is not present. This */
      /* will guarantee no change to state in the */
      /* instance where an endstate is not given. */
      all_states.states[state].inputs[countinputs].endstate=0xffffff;
      
      while(np2!=NULL)
      {
        /* scan through all subchildren */
        if(!xmlStrcmp(np2->name,(const xmlChar *)"perform"))
        {
          performstring=xmlNodeListGetString(doc,np2->xmlChildrenNode, 1);
          if(gnp_global_data.verbose_status)
            printf("    Perform string read: %s\n",performstring);

          all_states.states[state].inputs[countinputs].command=gnp_parse_command(performstring);
        }
        if(!xmlStrcmp(np2->name,(const xmlChar *)"endstate"))
        {
          lookup=gnp_lookup_state(xmlNodeListGetString(doc,np2->xmlChildrenNode, 1));
          if(lookup==0xffffff)
          {
            printf("ERROR: Invalid endstate for State %s, input key %s. ABORTING\n",
               all_states.nametable[state],thisname);
            exit(-1);
          }
          if(gnp_global_data.verbose_status)
            printf("    Endstate stored as %d\n",lookup);

          all_states.states[state].inputs[countinputs].endstate=lookup;
        }
        np2=np2->next;
      }
      countinputs++;
    }
    np1=np1->next;
  }

  return TRUE;
}


commands_struct *gnp_parse_command(gchar *command)

/* parse a string for all the commands it contains */
/* command strings have the format:                */
/*   <function>[(<param>[,...])][. ...]            */

{
  gchar *nextcommand,*commandline,*savestring;
  gchar *par1,*par2,*par3;
  guint  loc,base,lookup;
  commands_struct *cmdstr;

  if(command==NULL)
    return NULL;

  savestring=g_malloc(strlen(command));
  commandline=g_malloc(strlen(command));
  par1=g_malloc(strlen(command));
  par2=g_malloc(strlen(command));
  par3=g_malloc(strlen(command));
  cmdstr=g_malloc(sizeof(commands_struct));
  if(command[0]=='.')
    /* passed command starts with a '.'  */
    /* so we must strip it first         */
    strcpy(savestring,&command[1]);
  else
    strcpy(savestring,command);

  nextcommand=strpbrk(savestring,".\0");

  loc=strcspn(savestring,".(");

  strncpy(commandline,savestring,loc);
  commandline[loc]='\0';

  base=loc;
  if(savestring[base]=='(')
  {
    /* we have at least one parameter  */
    loc=strcspn(&savestring[base+1],",\")");
    if(savestring[base+loc+1]=='\"')
    {
      /* If we find a string, we must ignore  */
      /* finds of . and ) until the end of the*/
      /* string. gnp_scan_eos will return the */
      /* loc of the end-of-string.            */
      loc+=gnp_scan_eos(&savestring[base+loc+1]);
      /* we can't have another string in here */
      loc+=strcspn(&savestring[base+loc+1],",)");
    }    
    strncpy(par1,&savestring[base+1],loc);
    par1[loc]='\0';
    base+=loc+1;
    if(savestring[base]==',')
    {
      /* we have a second parameter */
      loc=strcspn(&savestring[base+1],",)");
      if(savestring[base+loc+1]=='\"')
      {
        /* If we find a string, we must ignore  */
        /* finds of . and ) until the end of the*/
        /* string. gnp_scan_eos will return the */
        /* loc of the end-of-string.            */
        loc+=gnp_scan_eos(&savestring[base+loc+1]);
        /* we can't have another string in here */
        loc+=strcspn(&savestring[base+loc+1],",)");
      }    
      strncpy(par2,&savestring[base+1],loc);
      par2[loc]='\0';
      base+=loc+1;
      if(savestring[base]==',')
      {
        /* we have a third parameter as well */
        loc=strcspn(&savestring[base+1],")");
        if(savestring[base+loc+1]=='\"')
        {
          /* If we find a string, we must ignore  */
          /* finds of . and ) until the end of the*/
          /* string. gnp_scan_eos will return the */
          /* loc of the end-of-string.            */
          loc+=gnp_scan_eos(&savestring[base+loc+1]);
          /* we can't have another string in here */
          loc+=strcspn(&savestring[base+loc+1],",)");
        }    
        strncpy(par3,&savestring[base+1],loc);
        par3[loc]='\0';
      }
      else
      {
        /* no third parameter */
        par3=NULL;
      }
    }
    else
    {
      /* no second and hence no third parameter */
      par2=NULL;
      par3=NULL;
    }
  }
  else
  {
    /* no parameters */
    par1=NULL;
    par2=NULL;
    par3=NULL;
  }

  if(gnp_global_data.verbose_status)
  {
    printf("      Command: %s, next command: %s\n",commandline, nextcommand);
    printf("       param1: %s, param2: %s, param3: %s\n",par1,par2,par3);
  }
  

  lookup=gnp_lookup_function_code_by_name(commandline);
  if(lookup==0xffffff)
  {
    printf("ERROR: Undefined function \"%s\". ABORTING\n",commandline);
    exit(-1);
  }
  cmdstr->function=lookup;
  if(par1!=NULL)
  {
    lookup=gnp_lookup_parameter(par1);
    if(lookup==0xffffff)
    {
      printf("ERROR: Undefined parameter \"%s\". ABORTING\n",par1);
      exit(-1);
    }
    cmdstr->param1=lookup;
  }
  else
    cmdstr->param1=0;
  if(par2!=NULL)
  {
    lookup=gnp_lookup_parameter(par2);
    if(lookup==0xffffff)
    {
      printf("ERROR: Undefined parameter \"%s\". ABORTING\n",par2);
      exit(-1);
    }
    cmdstr->param2=lookup;
  }
  else
    cmdstr->param2=0;
  if(par3!=NULL)
  {
    lookup=gnp_lookup_parameter(par3);
    if(lookup==0xffffff)
    {
      printf("ERROR: Undefined parameter \"%s\". ABORTING\n",par3);
      exit(-1);
    }
    cmdstr->param3=lookup;
  }
  else
    cmdstr->param3=0;

  cmdstr->next=gnp_parse_command(nextcommand);

  g_free(commandline);
  g_free(par1);
  g_free(par2);
  g_free(par3);

  return cmdstr;
}  


void  gnp_parse_function(xmlDocPtr *document, xmlNodePtr *node, guint count)

/*  Do all stuff related to the reading and setting up */
/*  of a function.                                     */

{


  function_table[count].param_count=0;
  function_table[count].command=NULL;
  function_table[count].sub=NULL;
  function_table[count].endstate=0;
}

guint gnp_scan_eos(gchar *string)

/* Scan a string until we find the      */
/* terminating ". Ignore quotes inside  */
/* the string that are escaped with the */
/*  /" sequence.                        */

{
  guint    loc=0;
  gboolean loop=TRUE;

  /* if this is not the start of a string */
  /* send it back unprocessed.            */
  if(string[0]!='\"')
    return 0;

  while(loop)
  {
    loc+=strcspn(&string[loc+1],"\"\0");
    if(string[loc]=='\0')
    {
      printf("ERROR: Unterminated string. ABORTING\n");
      exit(-1);
    }
    if(string[loc-1]!='\\')
      loop=FALSE;
  }

  return loc;
}


guint gnp_lookup_key(gchar *name)

/*  Lookup the key that matches the name  */
/* provided, and return that value.       */

{
  guint i=0;

  while(i<keysym_table_count)
  {
    if(strcmp(name,keysym_table[i].sym)==0)
      return keysym_table[i].code;
    i++;
  }
  /* if we get here, we have an undefined code */
  /* we return 0xffffff to flag this */
  return 0xffffff;
}


gchar *gnp_lookup_key_name(guint code)

/*  Lookup the key that matches the code  */
/* provided, and return that value.       */

{
  guint i=0;

  while(i<keysym_table_count)
  {
    if(keysym_table[i].code==code)
      return keysym_table[i].sym;
    i++;
  }
  /* if we get here, we have an undefined code */
  /* we return NULL to flag this */
  return NULL;
}


guint  gnp_lookup_state(gchar *name)

/* lookup the state name and return the    */
/* state array index that matches it.      */
/* Return 0xffffff if the state is not     */
/* defined to flag an error.               */

{
  guint i=0;

  while(i<all_states.max_states)
  {
    if(strcmp(name,all_states.nametable[i])==0)
      return i;
    i++;
  }
  /* if we get here, we have an undefined state */
  return 0xffffff;
}


gchar *gnp_lookup_state_name(guint code)

/* lookup the state code and return the    */
/* state name that matches it.             */
/* Return NULL if the state is not         */
/* defined to flag an error.               */

{
  if(code<all_states.max_states)
    return all_states.nametable[code];
  else
    return NULL;
}


guint  gnp_lookup_function_code_by_name(gchar *name)

/* lookup the provided function name and return  */
/* the function code that corresponds. If the    */
/* function is undefined, we return 0xffffff.    */

{
  guint i=0;

  while(i<function_table_count)
  {
    if(strcmp(name,function_table[i].sym)==0)
      return function_table[i].code;
    i++;
  }
  /* if we get here, we have an undefined code */
  /* we return 0xffffff to flag this */
  return 0xffffff;
}

guint  gnp_lookup_function(guint code)

/* lookup the provided function code, and return */
/* the table locate that corresponds.  If the    */
/* function code is undefined, return 0xffffffff */

{
  guint i=0;

  while(i<function_table_count)
  {
    if(function_table[i].code==code)
      return i;
    i++;
  }
  /* if we get here, we have an undefined code */
  /* we return 0xffffffff to flag this         */
  return 0xffffffff;
}


gchar *gnp_lookup_function_name(guint code)

/* lookup the provided function code and return  */
/* the function name that corresponds. If the    */
/* function is undefined, we return NULL.        */

{
  guint i=0;

  while(i<function_table_count)
  {
    if(function_table[i].code==code)
      return function_table[i].sym;
    i++;
  }
  /* if we get here, we have an undefined code */
  /* we return NULL to flag this               */
  return NULL;
}


guint  gnp_lookup_parameter(gchar *name)

/* lookup the provided param name and return     */
/* the param code that corresponds. If the       */
/* param is undefined, we return 0xffffff.       */

{
  guint i=0;

  while(i<register_table_count)
  {
    if(strcmp(name,register_table[i].sym)==0)
      return register_table[i].code;
    i++;
  }
  /* if we get here, we have an undefined code */
  /* we return 0xffffff to flag this */
  return 0xffffff;
}


gchar *gnp_lookup_parameter_name(guint code)

/* lookup the provided param code and return     */
/* the param name that corresponds. If the       */
/* param is undefined, we return NULL.           */

{
  guint i=0;

  while(i<register_table_count)
  {
    if(register_table[i].code==code)
      return register_table[i].sym;
    i++;
  }
  /* if we get here, we have an undefined code */
  /* we return NULL to flag this */
  return NULL;
}

